use futures::{future, Future, Poll, BoxFuture};
use futures::future::FutureResult;
use errors::HandleError;

/// Returned future from middlewares/handlers.
pub enum Outcome<T> {
    Result(FutureResult<T, HandleError>),
    Boxed(BoxFuture<T, HandleError>),
}

impl<T> Outcome<T> {
    pub fn from_future<F>(f: F) -> Self
    where
        F: Future<Item = T, Error = HandleError> + Send + 'static,
    {
        Outcome::Boxed(f.boxed())
    }

    pub fn continue_(ctx: T) -> Self {
        Outcome::Result(future::ok(ctx))
    }

    pub fn halt(err: HandleError) -> Self {
        Outcome::Result(future::err(err))
    }
}

impl<T: Send + 'static> Future for Outcome<T> {
    type Item = T;
    type Error = HandleError;

    #[inline(always)]
    fn poll(&mut self) -> Poll<Self::Item, Self::Error> {
        match *self {
            Outcome::Result(ref mut f) => f.poll(),
            Outcome::Boxed(ref mut f) => f.poll(),
        }
    }

    #[inline(always)]
    fn boxed(self) -> BoxFuture<Self::Item, Self::Error> {
        match self {
            Outcome::Result(f) => f.boxed(),
            Outcome::Boxed(f) => f,
        }
    }
}

#[macro_export]
macro_rules! try_f {
    ($e:expr) => (match $e {
        Ok(val) => val,
        Err(err) => return ::susanoo::Outcome::halt(
            ::susanoo::HandleError::Internal(Box::new(err))),
    });
}
