use std::borrow::Cow;
use hyper::header::{ContentType, Location};
use {RawResponse, StatusCode};
use request::Request;
use errors::HandleError;

/// [unstable] A response which contains all information to construct
pub struct Response {
    pub(crate) res: RawResponse,
}


pub trait Responder {
    fn respond(self, req: &Request) -> Result<RawResponse, HandleError>;
}

impl Responder for RawResponse {
    #[inline(always)]
    fn respond(self, _: &Request) -> Result<RawResponse, HandleError> {
        Ok(self)
    }
}

impl Responder for &'static str {
    fn respond(self, _: &Request) -> Result<RawResponse, HandleError> {
        Ok(
            RawResponse::new()
                .with_header(ContentType::plaintext())
                .with_body(self),
        )
    }
}

impl Responder for String {
    fn respond(self, _: &Request) -> Result<RawResponse, HandleError> {
        Ok(
            RawResponse::new()
                .with_header(ContentType::plaintext())
                .with_body(self),
        )
    }
}


pub struct Html<'a>(Cow<'a, str>);

impl<'a> Html<'a> {
    pub fn new<S: Into<Cow<'a, str>>>(body: S) -> Self {
        Html(body.into())
    }
}

impl<'a> Responder for Html<'a> {
    fn respond(self, _: &Request) -> Result<RawResponse, HandleError> {
        Ok(
            RawResponse::new()
                .with_header(ContentType::html())
                .with_body(self.0.into_owned()),
        )
    }
}


pub struct Redirect {
    status: StatusCode,
    uri: String,
}

impl Redirect {
    /// 302 Found
    pub fn found<S: Into<String>>(uri: S) -> Self {
        Redirect {
            status: StatusCode::Found,
            uri: uri.into(),
        }
    }
}

impl Responder for Redirect {
    fn respond(self, _: &Request) -> Result<RawResponse, HandleError> {
        Ok(
            RawResponse::new()
                .with_status(self.status)
                .with_header(Location::new(self.uri)),
        )
    }
}
