use std::sync::Arc;
use std::net::SocketAddr;

use futures::{Future, Stream};
use futures::future::BoxFuture;
use hyper::{Error as HyperError, Request as HyperRequest, Response as HyperResponse, StatusCode};
use hyper::server::{Http, Service};
use tokio_core::reactor::{Core, Handle};
use tokio_core::net::TcpListener;

use errors::{SusanooResult, HandleError};
use handler::Middleware;
use response::Response;
use router::{Router, Route};


/// Root instance of Web application.
#[derive(Default)]
pub struct Susanoo {
    listen_address: Option<SocketAddr>,
    router: Router,
    middlewares: Vec<Arc<Middleware>>,
}

impl Susanoo {
    /// Add a route to the application.
    pub fn with_route(mut self, route: Route) -> Self {
        self.router.add_route(route);
        self
    }

    /// Add a middleware to the application.
    ///
    /// If you want to apply middelewares to certain route, use `Route::with_middleware`.
    pub fn with_middleware<M: Middleware>(mut self, middleware: M) -> Self {
        self.middlewares.push(Arc::new(middleware));
        self
    }

    /// Set listen address of the application.
    ///
    /// The default address is `"0.0.0.0:4000"`.
    pub fn bind<S: AsRef<str>>(mut self, addr: S) -> SusanooResult<Self> {
        self.listen_address = Some(addr.as_ref().parse()?);
        Ok(self)
    }

    /// Run the server infinitely.
    pub fn run(self) -> SusanooResult<()> {
        let addr = self.listen_address
            .or_else(|| "0.0.0.0:4000".parse().ok())
            .unwrap();

        let mut core = Core::new()?;
        let handle = core.handle();

        let service = Arc::new(SusanooService {
            handle: handle.clone(),
            router: self.router,
            middlewares: self.middlewares,
        });

        let listener = TcpListener::bind(&addr, &handle)?;
        let server = listener.incoming().for_each(move |(sock, addr)| {
            Http::new().bind_connection(&handle, sock, addr, service.clone());
            Ok(())
        });

        core.run(server)?;
        Ok(())
    }
}


/// An asynchronous task executed by hyper.
pub struct SusanooService {
    handle: Handle,
    router: Router,
    middlewares: Vec<Arc<Middleware>>,
}

impl Service for SusanooService {
    type Request = HyperRequest;
    type Response = HyperResponse;
    type Error = HyperError;
    type Future = BoxFuture<HyperResponse, HyperError>;

    fn call(&self, req: HyperRequest) -> Self::Future {
        let base_middlewares = self.middlewares.iter();
        let remote = self.handle.remote().clone();
        let outcome = self.router.apply(req, remote, base_middlewares);

        outcome
            .then(|outcome| match outcome {
                Ok(Response { res }) => Ok(res),
                Err(HandleError::Raw(res)) => Ok(res),
                Err(HandleError::Status(status)) => {
                    // TODO: use custom error handler,
                    Ok(HyperResponse::new().with_status(status))
                }
                Err(HandleError::Internal(err)) => {
                    Ok(
                        HyperResponse::new()
                            .with_status(StatusCode::InternalServerError)
                            .with_body(format!("Internal server error: {:?}", err)),
                    )
                }
            })
            .boxed()
    }
}
